@@ -22,14 +22,14 @@ CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 import os
 import re
 import urllib
-from urlparse import unquote
 from collections import namedtuple
 from posixpath import normpath
 
 
 __all__ = ["ParseResult", "SplitResult", "parse", "extract", "split",
            "split_netloc", "split_host", "assemble", "encode", "normalize",
-           "normalize_path", "normalize_query"]
+           "normalize_host", "normalize_path", "normalize_query",
+           "normalize_fragment", "unquote"]
 
 
 PSL_URL = 'http://mxr.mozilla.org/mozilla-central/source/netwerk/dns/effective_tld_names.dat?raw=1'
@@ -54,9 +54,27 @@ PSL = _get_public_suffix_list()
 
 SCHEMES = ['http', 'https', 'ftp', 'sftp', 'file', 'gopher', 'imap', 'mms',
            'news', 'nntp', 'telnet', 'prospero', 'rsync', 'rtsp', 'rtspu',
-           'svn', 'git']
+           'svn', 'git', 'ws', 'wss']
 SCHEME_CHARS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
 IP_CHARS = '0123456789.:'
+DEFAULT_PORT = {
+    'http': '80',
+    'https': '443',
+    'ws': '80',
+    'wss': '443',
+    'ftp': '21',
+    'sftp': '22'
+}
+UNQUOTE_EXCEPTIONS = {
+    'path': ' /?+#',
+    'query': ' &=+#',
+    'fragment': ' +#'
+}
+
+_hextochr = {'%02x' % i: chr(i) for i in range(256)}
+_hextochr.update({'%02X' % i: chr(i) for i in range(256)})
+_idna_encode = lambda x: x.decode('utf-8').encode('idna')
+_idna_decode = lambda x: x.decode('idna').encode('utf-8')
 
 SplitResult = namedtuple('SplitResult', ['scheme', 'netloc', 'path', 'query',
                                          'fragment'])
@@ -68,34 +86,42 @@ ParseResult = namedtuple('ParseResult', ['scheme', 'username', 'password',
 def normalize(url):
     """Normalize a URL
     """
-    url_parts = split(url.strip())
-    if url_parts.scheme:
-        netloc = url_parts.netloc
-        path = url_parts.path
+    if url == '':
+        return ''
+    parts = split(url.strip())
+    if parts.scheme:
+        netloc = parts.netloc
+        if parts.scheme in SCHEMES:
+            path = normalize_path(parts.path)
+        else:
+            path = parts.path
     else:
-        netloc = url_parts.path
+        netloc = parts.path
         path = ''
         if '/' in netloc:
             tmp = netloc.split('/', 1)
             netloc = tmp[0]
-            path = '/' + tmp[1]
+            path = normalize_path('/' + tmp[1])
     username, password, host, port = split_netloc(netloc)
-    parts = ParseResult(url_parts.scheme, username, password, None, host, None,
-                        port, path, url_parts.query, url_parts.fragment)
-    return assemble(parts, default_path='/')
+    host = normalize_host(host)
+    port = normalize_port(parts.scheme, port)
+    query = normalize_query(parts.query)
+    fragment = normalize_fragment(parts.fragment)
+    result = ParseResult(parts.scheme, username, password, None, host, None,
+                         port, path, query, fragment)
+    return assemble(result)
 
 
 def encode(url):
     """Encode URL
     """
     parts = extract(url)
-    idna = lambda x: x.decode('utf-8').encode('idna')
-    encoded = ParseResult(*(idna(p) for p in parts))
+    encoded = ParseResult(*(_idna_encode(p) for p in parts))
     return assemble(encoded)
 
 
-def assemble(parts, default_path=''):
-    """Assemble a URL from the result returned by extract() or parse()
+def assemble(parts):
+    """Assemble a ParseResult to a new URL
     """
     nurl = ''
     if parts.scheme:
@@ -112,27 +138,44 @@ def assemble(parts, default_path=''):
     nurl += parts.domain
     if parts.tld:
         nurl += '.' + parts.tld
-    if parts.port and parts.port != '80':
+    if parts.port:
         nurl += ':' + parts.port
     if parts.path:
-        nurl += normalize_path(parts.path)
-    elif parts.scheme in SCHEMES:
-        nurl += default_path
+        nurl += parts.path
     if parts.query:
-        query = normalize_query(parts.query)
-        if query:
-            nurl += '?' + query
+        nurl += '?' + parts.query
     if parts.fragment:
         nurl += '#' + parts.fragment
     return nurl
 
 
+def normalize_host(host):
+    """Normalize host (decode IDNA)
+    """
+    if 'xn--' not in host:
+        return host
+    parts = host.split('.')
+    return '.'.join([_idna_decode(p) for p in parts])
+
+
+def normalize_port(scheme, port):
+    """Check if the port is default port
+    """
+    if not scheme:
+        return port
+    if port and port != DEFAULT_PORT[scheme]:
+        return port
+
+
 def normalize_path(path):
     """Normalize path (collapse etc.)
     """
-    if path in ['//', '/', '']:
+    if path in ['//', '/' ,'']:
         return '/'
-    return normpath(unquote(path))
+    npath = normpath(unquote(path, exceptions=UNQUOTE_EXCEPTIONS['path']))
+    if path[-1] == '/' and npath != '/':
+        npath += '/'
+    return npath
 
 
 def normalize_query(query):
@@ -140,7 +183,8 @@ def normalize_query(query):
     """
     if query == '' or len(query) <= 2:
         return ''
-    params = query.split('&')
+    nquery = unquote(query, exceptions=UNQUOTE_EXCEPTIONS['query'])
+    params = nquery.split('&')
     nparams = []
     for param in params:
         if '=' in param:
@@ -151,13 +195,37 @@ def normalize_query(query):
     return '&'.join(nparams)
 
 
+def normalize_fragment(fragment):
+    """Normalize fragment (unquote with exceptions only)
+    """
+    return unquote(fragment, UNQUOTE_EXCEPTIONS['fragment'])
+
+
+def unquote(text, exceptions=[]):
+    """Unquote a text but ignore the exceptions
+    """
+    if '%' not in text:
+        return text
+    s = text.split('%')
+    res = [s[0]]
+    for h in s[1:]:
+        c = _hextochr.get(h[:2])
+        if c and c not in exceptions:
+            if len(h) > 2:
+                res.append(c + h[2:])
+            else:
+                res.append(c)
+        else:
+            res.append('%' + h)
+    return ''.join(res)
+
+
 def parse(url):
     """Parse a URL
     """
     parts = split(url)
     if parts.scheme:
-        netloc = parts.netloc
-        (username, password, host, port) = split_netloc(netloc)
+        (username, password, host, port) = split_netloc(parts.netloc)
         (subdomain, domain, tld) = split_host(host)
     else:
         username = password = subdomain = domain = tld = port = ''
@@ -189,7 +257,10 @@ def split(url):
     """Split URL into scheme, netloc, path, query and fragment
     """
     scheme = netloc = path = query = fragment = ''
+    ip6_start = url.find('[')
     scheme_end = url.find(':')
+    if ip6_start > 0 and ip6_start < scheme_end:
+        scheme_end = -1
     if scheme_end > 0:
         for c in url[:scheme_end]:
             if c not in SCHEME_CHARS:
@@ -233,7 +304,7 @@ def split(url):
 
 
 def _clean_netloc(netloc):
-    """Remove trailing '.'s and ':'s from a URL and tolower
+    """Remove trailing '.' and ':' and tolower
     """
     return netloc.rstrip('.:').decode('utf-8').lower().encode('utf-8')
 
@@ -249,10 +320,8 @@ def split_netloc(netloc):
         else:
             username = user_pw
     netloc = _clean_netloc(netloc)
-    if '.' not in netloc:
-        return username, password, netloc, ''
-    if ':' in netloc:
-        host, port = netloc.split(':')
+    if ':' in netloc and netloc[-1] != ']':
+        host, port = netloc.rsplit(':', 1)
     else:
         host = netloc
     return username, password, host, port
